import dataclasses

import polars as pl
from pycomfort.files import *

import genotations.genomes

transcript_intersection = (set[str], (str, float, float)) #type alias for transcript intersections


@dataclasses.dataclass
class TranscriptIntersection:
    @staticmethod
    def extract_intervals(annotations: genotations.genomes.Annotations):
        """
        gets transcript interval, used for primers selection and other purposes
        :return:
        """
        return seq(annotations.annotations_df.with_column(annotations.coordinates_col).sort(pl.col("start"))\
                   .select([annotations.transcript_exon_col, pl.col("seqname"), pl.col("start"), pl.col("end")])\
                   .rows()).map(lambda row: TranscriptIntersection({row[0]}, row[1], row[2], row[3]))


    """
    TODO: move to primers module
    """
    transcripts: set[str]
    contig: str
    start: int
    end: int

    def to_graphical_feature(self):
        from dna_features_viewer import GraphicFeature
        return GraphicFeature(start=self.start, end=self.end, label=f"{self.contig}_inter_of_{len(self.transcripts)}", color=genotations.genomes.random_color())

    def length(self) -> int:
        return self.end - self.start

    def count(self) -> int:
        return len(self.transcripts)

    def merge(self, b: 'TranscriptIntersection') -> 'TranscriptIntersection':
        assert self.contig == b.contig, "to merge intervals contig should be the same"
        if self.transcripts.issubset(b.transcripts):
            return b
        elif b.transcripts.issubset(self.transcripts):
            return self
        else:
            start = max(self.start, b.start)
            end = min(self.end, b.end)
            joined: set[str] = self.transcripts.union(b.transcripts)
            return TranscriptIntersection(joined, self.contig, start, end)

    @staticmethod
    def find_deepest_intersection(intervals: list['TranscriptIntersection'], min_len: int = 20, previous: list['TranscriptIntersection'] = None):
        if len(intervals) < 2:
            return previous
        else:
            novel_intervals = seq(intervals) \
                .order_by(lambda ab: ab.start) \
                .sliding(2, 1) \
                .map(lambda ab: ab[0].merge(ab[1])) \
                .filter(lambda ab: ab.length() >= min_len)
            return TranscriptIntersection.find_deepest_intersection(novel_intervals.to_list(), min_len, intervals)

    @staticmethod
    def merge_intervals(intervals: list['TranscriptIntersection'], num: int = 1, min_len: int = 20):
        if num == 0 or len(intervals) < 2:
            return intervals
        else:
            novel_intervals = seq(intervals) \
                .order_by(lambda ab: ab.start) \
                .sliding(2, 1) \
                .map(lambda ab: ab[0].merge(ab[1])) \
                .filter(lambda ab: ab.length >= min_len)
            return TranscriptIntersection.merge_intervals(novel_intervals.to_list(), num - 1, min_len)

    @staticmethod
    def merge_interval_collection(intervals: list[transcript_intersection], min_len: int = 20, acc: list[list[transcript_intersection]] = None):
        if acc is None:
            return TranscriptIntersection.merge_interval_collection(intervals, min_len=min_len, acc=[intervals])
        if len(acc[-1]) <= 1:
            return acc
        else:
            acc.append(TranscriptIntersection.merge_intervals(intervals, len(acc), min_len))
            return TranscriptIntersection.merge_interval_collection(intervals, min_len, acc)



import dataclasses
import functools
from enum import Enum

import pandas
import polars as pl
from genomepy import Genome
import genomepy
from pycomfort.files import *
import random
from functools import cached_property, cache
from typing import Callable


class Strand(Enum):
    Plus = "+"
    Minus = "-"
    Undefined = ""

    def to_int(self) -> int:
        if self == Strand.Undefined:
            return 0
        elif self == Strand.Minus:
            return -1
        else:
            return 1

    def to_rc(self, feature_strand: str):
        """
        Converts to reverse_complement taking into consideration feature strand information
        :param feature_strand:
        :return:
        """
        return True if self == Strand.Minus or (self == Strand.Undefined and feature_strand == "-") else False



def _get_sequence_from_series(coordinates: Union[pl.Series, list], genome: Genome, strand: Strand = Strand.Undefined) -> str:
    """
    If the strand is undefined than
    :param coordinates:
    :param genome:
    :param strand:
    :return:
    """
    rc = strand.to_rc(coordinates[1])
    result = genome.get_seq(str(coordinates[0]), int(coordinates[2]), int(coordinates[3]), rc)
    return result.seq

def random_color():

    return "#"+''.join([random.choice('ABCDEF0123456789') for i in range(6)])


class FeatureType(Enum):
    STOP_CODON = 'stop_codon'
    CDS = 'CDS'
    THREE_PRIME_UTR = 'three_prime_utr'
    START_CODONG = 'start_codon'
    TRANSCRIPT = 'transcript'
    FIVE_PRIME_UTR = 'five_prime_utr'
    EXON = "exon"
    SELENOCYSTEINE = 'Selenocysteine'
    GENE = "gene"


class TranscriptBioType(Enum):
    RETAINED_INTRON = "retained_intron"
    PROTEIN_CODING = "protein_coding"
    PROTEIN_CODING_CDS_NOT_DEFINED = "protein_coding_CDS_not_defined"
    NONSENSE_MEDIATED_DECAY = "nonsense_mediated_decay"


class Annotations:
    """
    GTF annotations class,
    core class to work with GTF annotations in a chained way with polars.
    A widespread usage is calling chained methods and then getting resulting polars annotation_df dataframe
    """

    #declaration for the main polars dataframe in the class
    annotations_df: pl.DataFrame

    #this column expression is used in the functions that extract sequences. It just get's the fields important for sequence extraciton in one column

    coordinates_col: pl.Expr = pl.col("coordinates")
    coordinates_compute_col: pl.Expr = pl.concat_list([pl.col("seqname"), pl.col("strand"), pl.col("start"), pl.col("end")]).alias("coordinates")
    gene_col: pl.Expr = pl.col("gene")
    gene_name_col: pl.Expr = pl.col("gene_name")
    transcript_col: pl.Expr = pl.col("transcript")
    transcript_name_col: pl.Expr = pl.col("transcript_name")
    transcript_exon_col: pl.Expr = pl.col("transcript_exon")
    exon_col: pl.Expr = pl.col("exon")
    sequence_col: pl.Expr = pl.col("sequence")

    def __init__(self, gtf: Union[Path, str, pl.DataFrame]):
        """
        Major class constructor
        :param gtf: Path to GTF file as string or Path or polars Dataframe
        """
        if isinstance(gtf, Path) or type(gtf) is str:
            # if it is a path then load and clean it
            self.annotations_df = self.read_GTF(gtf)
        else:
            # if we already have the dataframe (for example in chained calls) than just assign it
            self.annotations_df = gtf

    def transform(self, fun: Callable[[pl.DataFrame], pl.DataFrame]):
        return Annotations(fun(self.annotations_df))

    def sort_by_transcript_exon(self):
        return self.annotations_df.sort([pl.col(self.transcript_name_col), self.exon_col])

    @cache
    def read_GTF(self, path: Union[str, Path]) -> pl.DataFrame:
        """
        Reads GTF file with feature annotations.
        It also does preprocessing of it as often in GTF multiple values are mixed in one cell
        :param path: path to the file
        :return: polards datafra
        """
        att = pl.col("attribute")
        loaded = pl.read_csv(str(path), has_header=False, comment_char="#", sep="\t",
                             new_columns=["seqname", "source", "feature", "start", "end", "score", "strand", "frame", "attribute"],
                             dtypes={
                                 #sometimes polars makes mistakes in automatic type derivation with some fields
                                 "seqname": pl.Utf8, #pl.Categorical,
                                 "start": pl.UInt64, "end": pl.UInt64,
                                 "strand": pl.Categorical
                             }
                             )
        transcript_exon_compute = (self.transcript_name_col+pl.lit("_")+pl.col("exon_number")).alias("transcript_exon")
        # does some preprocessing, mostly extracting attributes (which are multiple per cell in GTF files) to separate columns with regular expressions
        result = loaded \
                 .with_column(att.str.extract("gene_id \"[a-zA-Z0-9_.-]*", 0).str.replace("gene_id \"", "").alias("gene")) \
                 .with_column(att.str.extract("gene_name \"[a-zA-Z0-9_.-]*", 0).str.replace("gene_name \"", "").alias("gene_name")) \
                 .with_column(att.str.extract("gene_biotype \"[a-zA-Z0-9_.-]*", 0).str.replace("transcript_biotype \"", "").alias("transcript_biotype")) \
                 .with_column(att.str.extract("transcript_id \"[a-zA-Z0-9_.-]*", 0).str.replace("transcript_id \"", "").alias("transcript")) \
                 .with_column(att.str.extract("transcript_name \"[a-zA-Z0-9_.-]*", 0).str.replace("transcript_name \"", "").alias("transcript_name")) \
                 .with_column(att.str.extract("transcript_biotype \"[a-zA-Z0-9_.-]*", 0).str.replace("transcript_biotype \"", "").alias("transcript_biotype")) \
                 .with_column(att.str.extract("exon_id \"[a-zA-Z0-9_.-]*", 0).str.replace("exon_id \"", "").cast(pl.Utf8).alias("exon")) \
                 .with_column(att.str.extract("exon_number \"[0-9_.-]*", 0).str.replace("exon_number \"", "").cast(pl.UInt64).alias("exon_number")) \
                 .with_column(transcript_exon_compute)
        return result


    def with_coordinates_column(self) -> 'Annotations':
        return self if "coordinates" in self.annotations_df.columns else Annotations(self.annotations_df.with_column(self.coordinates_compute_col))

    def has_sequence(self) -> bool:
        return "sequence" in self.annotations_df.columns

    def _optional_sequence(self, selection: list[pl.col]) -> list[pl.col]:
        return selection + [self.sequence_col] if self.has_sequence() else selection

    def with_genes_only(self) -> 'Annotations':
        return Annotations(self.genes().annotations_df.select(self._optional_sequence([self.gene_col, self.gene_name_col])))

    def with_genes_coordinates_only(self) -> 'Annotations':
        to_select = self._optional_sequence([self.gene_col, self.gene_name_col, self.coordinates_col])
        return Annotations(self.genes().with_coordinates_column().annotations_df.select(to_select))

    def with_genes_transcripts_only(self) -> 'Annotations':
        to_select = self._optional_sequence([self.gene_col, self.gene_name_col, self.transcript_col, self.transcript_name_col])
        return Annotations(self.transcripts().annotations_df.select(to_select))

    def with_genes_transcripts_coordinates_only(self) -> 'Annotations':
        to_select = self._optional_sequence([self.gene_col, self.gene_name_col, self.transcript_col, self.transcript_name_col, self.coordinates_col])
        return Annotations(self.transcripts().with_coordinates_column().annotations_df.select(to_select))

    def with_genes_transcripts_exons_coordinates_only(self):
        """
        TODO: rename properly, core idea is just to keep only essential fields
        :return:
        """
        to_select = self._optional_sequence([self.gene_col, self.gene_name_col,
                     self.transcript_col, self.transcript_name_col, self.exon_col, self.transcript_exon_col,
                     self.coordinates_col])
        return Annotations(self.exons().with_coordinates_column().annotations_df.select(to_select))

    def extend_with_annotations(self, expressions: pl.DataFrame):
        cols = expressions.columns
        assert "transcript" in cols or "gene" in cols, "expresison dataframe has to have either transcript or gene column"
        by = self.transcript_col if "transcript" in cols else self.gene_col
        return Annotations(self.annotations_df.join(expressions, on=by)).annotations_df

    def extend_with_annotations_and_sequences(self, expressions: pl.DataFrame, genome: Genome, strand: Strand = Strand.Plus):
        return Annotations(self.extend_with_annotations(expressions)).with_sequences(genome, strand).annotations_df

    @cached_property
    def annotations_pandas(self) -> pandas.DataFrame:
        """
        For strange people who prefer reading results as pandas instead of polars
        :return:
        """
        return self.annotations_df.to_pandas()

    def with_gene_name_in(self, *genes) -> 'Annotations':
        """
        Keeps only features which contains specified string in their gene_name
        :param gene_name: part of gene_name to search for, WARNING: case sensitive!
        :return: Self with filtered annotation_df dataframe for further chained calls
        """
        result = self.annotations_df \
            .filter(pl.col("gene_name").is_in(genes))
        return Annotations(result)

    def with_transcript_name_in(self, *transcripts) -> 'Annotations':

        result = self.annotations_df \
            .filter(pl.col("transcript_name").is_in(transcripts))
        return Annotations(result)

    def with_gene_name_contains(self, gene_name: str) -> 'Annotations':
        """
        Keeps only features which contains specified string in their gene_name
        :param gene_name: part of gene_name to search for, WARNING: case sensitive!
        :return: Self with filtered annotation_df dataframe for further chained calls
        """
        result = self.annotations_df \
                 .filter(pl.col("gene_name").str.contains(gene_name))
        return Annotations(result)

    def by_gene_id(self, gene_id: str) -> 'Annotations':
        """
        If we know Ensembl gene ID and want to get only its features
        :param gene_id:
        :return: self with filtered annotation_df for further chained calls
        """
        result = self.annotations_df \
                 .filter(pl.col("gene").str.contains(gene_id))
        return Annotations(result)

    def by_transcript_name(self, transcript_name: str) -> 'Annotations':
        """
        If we know Ensembl transcript ID and want to get only its features
        :param transcript_name:
        :return: self with filtered annotation_df for further chained calls
        """
        result = self.annotations_df \
                 .filter(pl.col("transcript_name").str.contains(transcript_name)).unique()
        return Annotations(result)

    def by_transcript_id(self, transcript_id: str) -> 'Annotations':
        result = self.annotations_df \
                 .filter(pl.col("transcript").str.contains(transcript_id))
        return Annotations(result)

    def protein_coding(self) -> 'Annotations':
        """
        Filteres only transcript features that are protein-coding
        :return: self with filtered annotation_df for further chained calls
        """
        result = self.annotations_df.filter(pl.col("transcript_biotype").str.contains(TranscriptBioType.PROTEIN_CODING.value))
        return Annotations(result)

    def features(self, features: list[str]) -> 'Annotations':
        """
        Filters by feature types we are interested in
        :param features: list of feature types we are interested in
        :return: self with filtered annotation_df for further chained calls
        """
        result = self.annotations_df.filter(pl.col("feature").is_in(features))
        return Annotations(result)

    def feature(self, feature: FeatureType) -> 'Annotations':
        """
        Filters by feature type selected from Enum options
        :param feature: FeatureType enum value
        :return: self with filtered annotation_df for further chained calls
        """
        result = self.annotations_df.filter(pl.col("feature") == feature.value)
        return Annotations(result)

    def exons(self) -> 'Annotations':
        """
        gets only exons
        :return: self with filtered annotation_df for further chained calls
        """
        return self.feature(FeatureType.EXON)

    def transcripts(self) -> 'Annotations':
        return self.feature(FeatureType.TRANSCRIPT)


    def genes(self) -> 'Annotations':
        """
        Gets only genes features
        :return:
        """
        return self.feature(FeatureType.GENE)

    @cached_property
    def gene_names_df(self) -> pl.DataFrame:
        """
        Getting just Ensembl gene id -> gene name correspondence is a common task as we often want to join such dataframe with others
        :return: NOTE: returns dataframe and not self
        """
        return self.genes().annotations_df.select([pl.col("gene"), pl.col("gene_name")]).unique()

    def _strings_to_spans(self, strings: list[str]):
        return seq(strings).fold_left( ((0, 0),), lambda acc, el: acc + ((acc[-1][0]+acc[-1][1], len(el)),) ).to_list()[1:]

    def get_transcript_sequences(self, genome: genomepy.genome = None, strand: Strand = Strand.Undefined) -> pl.DataFrame:
        assert genome is not None or self.has_sequence(), "there should be either sequence or genome available!"
        if not self.has_sequence():
            return self.with_sequences(genome, strand).get_transcript_sequences()
        else:
            return self.annotations_df.sort([self.transcript_name_col, pl.col("exon_number")])\
                .groupby(self.transcript_name_col, maintain_order=True)\
                .agg([self.sequence_col])\
                .with_column(pl.col("sequence").apply(self._strings_to_spans).alias("spans"))\
                .with_column(pl.col("sequence").apply(lambda r: seq(r).reduce(lambda a,b : a+b)).alias("mRNA"))

    @cached_property
    def transcript_gene_names_df(self) -> pl.DataFrame:
        """
        Getting just  Ensembl Transcript id -> Transcript name and Ensembl gene id -> gene name correspondence is a common task as we often want to join such dataframe with others
        :return: NOTE: returns dataframe and not self, _df suffix means dataframe
        """
        return self.annotations_df\
            .select([pl.col("transcript"), pl.col("transcript_name"), pl.col("gene"), pl.col("gene_name")])\
            .filter(pl.col("transcript").is_not_null())\
            .unique()

    def get_transcript_ids(self) -> pl.Series:
        """
        :return: a series with all transcript ids in the annotations
        """
        return self.annotations_df.select(pl.col("transcript_id")).to_series()

    def get_transcript_names(self) -> pl.Series:
        """
        :return: a series with all transcript ids in the annotations
        """
        return self.annotations_df.select(pl.col("transcript_name").unique()).to_series()

    def between(self, start: int, end: int) -> 'Annotations':
        return Annotations(self.annotations_df.filter(pl.col("start") >= start & pl.col("end") <= end))


    def exon_features_by_gene_name(self, gene_name: str) -> seq:
        """
        Visualizing exon gene features, uses dna_feature_viewer library for drawing
        :param gene_name:
        :return:
        """
        from dna_features_viewer import GraphicFeature
        anno = self.with_gene_name_contains(gene_name).protein_coding().exons().\
            annotations_df.select(["transcript_exon", "start", "end"]).unique()
        return seq(anno.rows()) \
                .map(lambda t:  GraphicFeature(
                start = t[1],
                end=t[2],
                label=t[0],
                open_left=True,
                open_right=True,
                color=random_color())
            ) \
            .to_list()

    def transcript_features_by_gene_name(self, gene_name, strand: Strand = Strand.Plus) -> list:
        """
        Visualizing exon gene features, uses dna_feature_viewer library for drawing
        :param gene_name: part of the gene name of interest
        :return: list of graphical features that then can be rendered to plots
        """
        from dna_features_viewer import GraphicFeature
        transcripts_for_gene = self.with_gene_name_contains(gene_name).transcripts()
        return seq(transcripts_for_gene.annotations_df.select(
            ["transcript_name", "strand", "start", "end"]).rows()) \
                .map(lambda t:  GraphicFeature(
                    start=t[2] if not strand.to_rc(t[1]) else t[3],
                    end=t[3] if not strand.to_rc(t[1]) else t[2],
                    label=t[0],
                    open_left=True,
                    open_right=True,
                    color=random_color(),
                    strand=1 if not strand.to_rc(t[1]) else -1
                )
            )\
            .to_list()

    def _gene_to_graphical_record(self, gene_name: str, gene_strand: str, start: int, end: int,
                                 sequence: str, exons: bool = True,
                                 transcript_intersections: list['TranscriptIntersection'] = None,
                                 other_features: list = None, strand: Strand = Strand.Plus
                                 ):
        """
        Writes a graphical record from the gene with additional parameters, is considered protected function
        :param gene_name: part of the gene name of interest
        :param start:
        :param end:
        :param sequence:
        :param exons:
        :param transcript_intersections:
        :param other_features:
        :param strand:
        :return: GraphicalRecord that then can be rendered as plot by calling plot() method
        """
        strand = -1 if strand == Strand.Minus else Strand.Plus
        from dna_features_viewer import GraphicRecord, GraphicFeature
        rc = strand.to_rc(gene_strand)
        source = GraphicFeature(start=start if not rc else end, end=end if not rc else start, label=gene_name, open_left=True, open_right=True, strand=strand.to_int())
        features = self.exon_features_by_gene_name(gene_name) if exons else self.transcript_features_by_gene_name(gene_name, strand)
        intersection_features = [] if transcript_intersections is None else seq(transcript_intersections).map(lambda t: t.to_graphical_feature()).to_list()
        other = [] if other_features is None else other_features
        features = [source] + features + intersection_features + other
        return GraphicRecord(sequence=sequence, first_index=start, features=features)


    def genes_visual(self, genome: Genome, strand: Strand = Strand.Plus, exons: bool = True,
                     transcript_intersections: list['TranscriptIntersection'] = None,
                     other_features: list = None):
        """
        visualizes genes with dna_feature_viewer
        :param genome: genomepy genome that will be used to extract sequences
        :param strand: strand that we want to see
        :param exons: if we include exons in the render
        :param transcript_intersections:
        :param other_features:
        :return:
        """
        annotation_with_sequence = self.genes().with_sequences(genome, strand).annotations_df
        return seq(
                annotation_with_sequence.select(["gene_name", "strand", "start", "end", "sequence"]).rows()
            ).map(lambda r: self._gene_to_graphical_record(
                gene_name=r[0], gene_strand=r[1], start=r[2], end=r[3], sequence=r[4],
                exons=exons, strand=strand,
                transcript_intersections=transcript_intersections,
                other_features=other_features
                )
            ).to_list()

    def exons_by_transcript_name(self, transcript_name: str) -> 'Annotations':
        """
        :param transcript_name: transcript from which to take exons
        :return: self with filtered annotation_df for further chained calls
        """
        return Annotations(self.by_transcript_name(transcript_name).exons().annotations_df.sort(pl.col("exon_number")))

    def with_sequences(self, genome: Genome, strand: Strand = Strand.Undefined) -> 'Annotations':
        """
        extends the annotations with sequences
        :param genome: genomepy genome instance
        :param strand: if Undefined - gets strand from features info and applied reverse complement if it is minus
        :return:
        """
        if "sequence" in self.annotations_df.columns:
            print("sequence column already exists, no work needed!")
            return self
        else:
            if self.annotations_df.shape[0] > 100:
                print(f"There are {self.annotations_df.shape} annotations, loading sequences can take quite a while!")
            extract_sequence = functools.partial(_get_sequence_from_series, genome=genome, strand=strand)
            with_sequences = self.with_coordinates_column().annotations_df.with_column(self.coordinates_col.apply(extract_sequence).alias("sequence"))
            return Annotations(with_sequences)

    def get_intervals_with_set(self):
        """
        gets interval sets, used for primers selection and other purposes
        :return:
        """
        return self.annotations_df.with_column(self.coordinates_col) \
            .select([pl.col("seqname"), pl.col("start"), pl.col("end")]).distinct().apply(lambda r: (set(r[0]), r[1], r[2])).rows()



import re
from copy import copy
from enum import Enum

import Bio
import genomepy
from Bio import Seq
import pyfaidx
from primer3 import bindings
import random
#from cloning.features import *
from typing import *
from functional import seq
from dna_features_viewer import GraphicFeature, GraphicRecord
from dataclasses import dataclass

import genotations.genomes
from genotations.genomes import Strand
from genotations.genomes import random_color

from genotations.intersections import TranscriptIntersection
import polars as pl


class PrimersMode(Enum):
    FLANK = 1 #it should flank the part
    PART_LEFT_JUNCTION_LEFT = 2 # from left part to junction of the tested feature
    JUNCTION_RIGHT_PART_RIGHT =3 # takes junction of the part to its right and the rest is inside leftish parts
    TWO_PARTS_JUNCTIONS = 4 # takes junction of the part and junction of the part to the right


@dataclass
class PrimerResult:
    PRIMER_PAIR_PENALTY: float
    PRIMER_LEFT_PENALTY: float
    PRIMER_RIGHT_PENALTY: float
    #PRIMER_INTERNAL_PENALTY: float
    PRIMER_LEFT_SEQUENCE: str
    PRIMER_RIGHT_SEQUENCE: str
    #PRIMER_INTERNAL_SEQUENCE: str
    PRIMER_LEFT: tuple[int, int]
    PRIMER_RIGHT: tuple[int, int]
    #PRIMER_INTERNAL: tuple[int, int]
    PRIMER_LEFT_TM: float
    PRIMER_RIGHT_TM: float
    #PRIMER_INTERNAL_TM: float
    PRIMER_LEFT_GC_PERCENT: float
    PRIMER_RIGHT_GC_PERCENT: float
    #PRIMER_INTERNAL_GC_PERCENT: float
    PRIMER_LEFT_SELF_ANY_TH: float
    PRIMER_RIGHT_SELF_ANY_TH: float
    #PRIMER_INTERNAL_SELF_ANY_TH: float
    #PRIMER_LEFT_SELF_END_TH: float
    #PRIMER_RIGHT_SELF_END_TH: float
    #PRIMER_INTERNAL_SELF_END_TH: float
    PRIMER_LEFT_HAIRPIN_TH: float
    PRIMER_RIGHT_HAIRPIN_TH: float
    #PRIMER_INTERNAL_HAIRPIN_TH: float
    PRIMER_LEFT_END_STABILITY: float
    PRIMER_RIGHT_END_STABILITY: float
    PRIMER_PAIR_COMPL_ANY_TH: float
    PRIMER_PAIR_COMPL_END_TH: float
    PRIMER_PAIR_PRODUCT_SIZE: int

    def has_in_any_of_primers(self, sequence: str):
        return sequence in self.PRIMER_LEFT_SEQUENCE or sequence in self.PRIMER_RIGHT_SEQUENCE

    def get_sequences(self):
        return (self.PRIMER_LEFT_SEQUENCE, self.PRIMER_RIGHT_SEQUENCE)

    def counts(self, sequence: Bio.Seq.Seq) -> tuple[int, int]:
        rev = sequence.reverse_complement()
        return sequence.count_overlap(self.PRIMER_LEFT_SEQUENCE) + rev.count_overlap(self.PRIMER_LEFT_SEQUENCE), \
               rev.count_overlap(self.PRIMER_RIGHT_SEQUENCE) + sequence.count_overlap(self.PRIMER_RIGHT_SEQUENCE)

    def has_repeats(self, sequence: Bio.Seq.Seq) -> bool:
        (f, r) = self.counts(sequence)
        return f + r > 2

    def get_temperatures(self):
        return self.PRIMER_LEFT_TM, self.PRIMER_RIGHT_TM

    def check_temperature(self, max_diff: float = 3.0):
        return abs(self.PRIMER_LEFT_TM-self.PRIMER_RIGHT_TM) <= max_diff

    def get_hairpins(self):
        return self.PRIMER_LEFT_HAIRPIN_TH, self.PRIMER_RIGHT_HAIRPIN_TH

    def start_with_g(self):
        return bool(re.search("^G", self.PRIMER_LEFT_SEQUENCE) or re.search("^G", self.PRIMER_RIGHT_SEQUENCE))

    def has_n_repeated_gc(self, num: int) -> bool:
        return bool(re.search('[G|C]' * num, self.PRIMER_LEFT_SEQUENCE) or
                    bool(re.search('[G|C]' * num, self.PRIMER_RIGHT_SEQUENCE)))

    def one_two_three_gc_in_last_four(self):
        gc = self.PRIMER_LEFT_SEQUENCE[-4:].count("G") + self.PRIMER_LEFT_SEQUENCE[-4:].count("C")
        return 1 < gc < 4

    def half_gc_in_last_four(self):
        return (self.PRIMER_LEFT_SEQUENCE[-4:].count("G") + self.PRIMER_LEFT_SEQUENCE[-4:].count("C")) == 2

    def filter(self, max_repeated_gc: int = 4,
               max_temperature_difference: float = 3,
               not_start_with_g: bool = False,
               one_two_three_gc_in_last_four: bool = True,
               only_half_gc_in_last_four: bool = False):
        if self.has_n_repeated_gc(max_repeated_gc):
            return False
        elif not self.check_temperature(max_temperature_difference):
            return False
        elif not_start_with_g and self.start_with_g():
            return False
        elif one_two_three_gc_in_last_four and not self.one_two_three_gc_in_last_four():
            return False
        elif only_half_gc_in_last_four and not self.half_gc_in_last_four():
            return False
        else:
            return True


    @staticmethod
    def extract(dic: Dict, num: int) -> 'PrimerResult':
        return PrimerResult(
            PRIMER_PAIR_PENALTY = dic[f"PRIMER_PAIR_{num}_PENALTY"],
            PRIMER_LEFT_PENALTY = dic[f"PRIMER_LEFT_{num}_PENALTY"],
            PRIMER_RIGHT_PENALTY = dic[f"PRIMER_RIGHT_{num}_PENALTY"],
            #PRIMER_INTERNAL_PENALTY = dic[f"PRIMER_INTERNAL_{num}_PENALTY"],
            #PRIMER_INTERNAL_PENALTY = dic[f"PRIMER_INTERNAL_{num}_PENALTY"],
            PRIMER_LEFT_SEQUENCE = dic[f"PRIMER_LEFT_{num}_SEQUENCE"],
            PRIMER_RIGHT_SEQUENCE = dic[f"PRIMER_RIGHT_{num}_SEQUENCE"],
            #PRIMER_INTERNAL_SEQUENCE = dic[f"PRIMER_INTERNAL_{num}_SEQUENCE"],
            PRIMER_LEFT = dic[f"PRIMER_LEFT_{num}"],
            PRIMER_RIGHT = dic[f"PRIMER_RIGHT_{num}"],
            #PRIMER_INTERNAL = dic[f"PRIMER_INTERNAL_{num}"],
            PRIMER_LEFT_TM = dic[f"PRIMER_LEFT_{num}_TM"],
            PRIMER_RIGHT_TM = dic[f"PRIMER_RIGHT_{num}_TM"],
            #PRIMER_INTERNAL_TM = dic[f"PRIMER_INTERNAL_{num}_TM"],
            PRIMER_LEFT_GC_PERCENT = dic[f"PRIMER_LEFT_{num}_GC_PERCENT"],
            PRIMER_RIGHT_GC_PERCENT = dic[f"PRIMER_RIGHT_{num}_GC_PERCENT"],
            #PRIMER_INTERNAL_GC_PERCENT = dic[f"PRIMER_INTERNAL_{num}_GC_PERCENT"],
            PRIMER_LEFT_SELF_ANY_TH = dic[f"PRIMER_LEFT_{num}_SELF_ANY_TH"],
            PRIMER_RIGHT_SELF_ANY_TH = dic[f"PRIMER_RIGHT_{num}_SELF_ANY_TH"],
            #PRIMER_INTERNAL_SELF_ANY_TH = dic[f"PRIMER_INTERNAL_{num}_SELF_ANY_TH"],
            #PRIMER_LEFT_SELF_END_TH = dic[f"PRIMER_INTERNAL_{num}_SELF_ANY_TH"],
            #PRIMER_RIGHT_SELF_END_TH = dic[f"PRIMER_RIGHT_{num}_SELF_END_TH"],
            #PRIMER_INTERNAL_SELF_END_TH= dic[f"PRIMER_INTERNAL_{num}_SELF_END_TH"],
            PRIMER_LEFT_HAIRPIN_TH = dic[f"PRIMER_LEFT_{num}_HAIRPIN_TH"],
            PRIMER_RIGHT_HAIRPIN_TH = dic[f"PRIMER_RIGHT_{num}_HAIRPIN_TH"],
            #PRIMER_INTERNAL_HAIRPIN_TH = dic[f"PRIMER_INTERNAL_{num}_HAIRPIN_TH"],
            PRIMER_LEFT_END_STABILITY = dic[f"PRIMER_LEFT_{num}_END_STABILITY"],
            PRIMER_RIGHT_END_STABILITY = dic[f"PRIMER_RIGHT_{num}_END_STABILITY"],
            PRIMER_PAIR_COMPL_ANY_TH = dic[f"PRIMER_PAIR_{num}_COMPL_ANY_TH"],
            PRIMER_PAIR_COMPL_END_TH = dic[f"PRIMER_PAIR_{num}_COMPL_END_TH"],
            PRIMER_PAIR_PRODUCT_SIZE = dic[f"PRIMER_PAIR_{num}_PRODUCT_SIZE"]
        )


class PrimerResults:

    results: list[PrimerResult]
    primer_left_explain: str
    primer_right_explain: str

    @staticmethod
    def empty() -> 'PrimerResults':
        return PrimerResults({})

    def __init__(self, dic: Dict[str, any]):
        nums = self.get_nums(dic)
        self.results: list[PrimerResult] = [PrimerResult.extract(dic, i) for i in nums]
        self.primer_left_explain: str = dic["PRIMER_LEFT_EXPLAIN"] if dic != {} else None
        self.primer_right_explain: str = dic["PRIMER_RIGHT_EXPLAIN"] if dic != {} else None
        #self.primer_internal_explain: str = dic["PRIMER_INTERNAL_EXPLAIN"] if dic != {} else None

    def get_nums(self, dic: Dict[str, any]):
        return [int(key[key.index("LEFT_")+5:]) for key in dic.keys() if key.startswith("PRIMER_LEFT_") and key.count("_") == 2 and "EXPLAIN" not in key]

    def get_cleaned_results(self, max_repeated_gc: int = 4,
                            max_temperature_difference: float = 3,
                            not_start_with_g: bool = False,
                            one_two_three_gc_in_last_four: bool = True,
                            only_half_gc_in_last_four: bool = False):
        return [r for r in self.results if r.filter(max_repeated_gc, max_temperature_difference, not_start_with_g, one_two_three_gc_in_last_four, only_half_gc_in_last_four)]

    def not_repeated_in(self, bio: Bio.Seq):
        new_results = seq(self.results).filter(lambda r: not r.has_repeats(bio)).to_list()
        result = copy(self)
        result.results = new_results
        return result

def suggest_primers_for_transcript_by_exons(anno: genotations.genomes.Annotations, transcript_name: str,
    from_exon: int, to_exon: int,
    opt_t: float = 60.0,
    min_t: float = 59.5,
    max_t: float = 64.0,
    min_product: int = 80,
    max_product: int = 220,
    genome: genotations.genomes.Genome = None,
    debug: bool = False) -> list[PrimerResults]:
    df = anno.with_transcript_name_in(transcript_name).get_transcript_sequences(genome)
    return suggest_primers_by_exons(df, from_exon, to_exon, opt_t, min_t, max_t, min_product, max_product, debug)[transcript_name]

def suggest_primers_by_exons(df: pl.DataFrame, from_exon: int, to_exon: int,
                        opt_t: float = 60.0,
                        min_t: float = 59.5,
                        max_t: float = 64.0,
                        min_product: int = 80,
                        max_product: int = 220,
                        debug: bool = False) -> OrderedDict[str, list[PrimerResults]]:

    def apply_suggestion(row: list[str]):
        [transcript_name, sequence, spans, mRNA] = row
        from_span = spans[from_exon-1]
        to_span = spans[to_exon-1]
        primers_pair_ok = (from_span[0], from_span[1], to_span[0], to_span[1])
        return (transcript_name, suggest_primers(mRNA, opt_t, min_t, max_t,
                               primers_pair_ok = primers_pair_ok, debug=debug,
                               min_product=min_product, max_product=max_product))
    """
    assuming ['transcript_name', 'sequence', 'spans', 'mRNA'] structure
    :param df:
    :return:
    """
    assert from_exon < to_exon, "should be from smaller exon to larger for the sake of clarity"
    return OrderedDict( seq(df.rows()).map(apply_suggestion).filter(lambda v:  len(v[1].results)>0).to_list() )



def suggest_primers(dna: str, opt_t: float = 60.0,
                    min_t: float = 57.0,  max_t: float = 62.0,
                    sequence_id: str = "sequence_for_primers",
                    primers_pair_ok: tuple[int, int, int, int] = None,
                    min_product: int = 30,
                    max_product: int = 3000,
                    debug: bool = False) -> PrimerResults:
    sequence_dict: Dict[str, any] = {
        'SEQUENCE_ID': sequence_id,
        'SEQUENCE_TEMPLATE': dna
    }
    primer_dict: Dict[str, any] = {
        'PRIMER_OPT_SIZE': 20,
        #'PRIMER_INTERNAL_MAX_SELF_END': 8,
        'PRIMER_MIN_SIZE': 18,
        'PRIMER_MAX_SIZE': 25,
        'PRIMER_OPT_TM': opt_t, #60 is optimal for rt qpcr, 58 for normal
        'PRIMER_MIN_TM': min_t,
        'PRIMER_MAX_TM': max_t,
        'PRIMER_MIN_GC': 20.0,
        'PRIMER_MAX_GC': 80.0,
        'PRIMER_MAX_POLY_X': 100,
        #'PRIMER_INTERNAL_MAX_POLY_X': 100,
        'PRIMER_SALT_MONOVALENT': 50.0,
        'PRIMER_DNA_CONC': 50.0,
        'PRIMER_MAX_NS_ACCEPTED': 0,
        'PRIMER_MAX_SELF_ANY': 12,
        'PRIMER_MAX_SELF_END': 8,
        'PRIMER_PAIR_MAX_COMPL_ANY': 12,
        'PRIMER_PAIR_MAX_COMPL_END': 8,
        'PRIMER_PRODUCT_SIZE_RANGE': [min_product, max_product]
    }
    if primers_pair_ok is not None:
        sequence_dict["SEQUENCE_PRIMER_PAIR_OK_REGION_LIST"] = [primers_pair_ok[0], primers_pair_ok[1], primers_pair_ok[2], primers_pair_ok[3]]

    result_dict: Dict[str, any] = bindings.designPrimers(sequence_dict, primer_dict) #,'PRIMER_PRODUCT_SIZE_RANGE': ranges #500 20-25 #<3kb
    result = PrimerResults(result_dict)
    if debug:
        cleaned_results = result.get_cleaned_results()
        print(f"""found {len(result.results)} primers,
        out of which {len(cleaned_results)} should be more or less ok
        spans of the primers are: {[(v.PRIMER_LEFT[0], v.PRIMER_LEFT[1]) for v in cleaned_results]}
        out of which {len(result.get_cleaned_results(max_repeated_gc=5))} should have relaxed (5) max repeated GC
        out of which {len(result.get_cleaned_results(not_start_with_g=True))} should also not start with G
        out of which {len(result.get_cleaned_results(only_half_gc_in_last_four=True))} should also have half gc
        """)
    return PrimerResults(result_dict)


def suggest_interval_primers(s: pyfaidx.Sequence,
                             a: TranscriptIntersection, b: TranscriptIntersection,
                             opt_t: float = 60.0,
                             min_t: float = 59.9,
                             max_t: float = 65.0) -> PrimerResults:
    a_left = a.start - s.start
    a_len = a.end - a.start
    b_left = b.start - s.start
    b_len = b.end - b.start
    primers_pair_ok = (a_left, a_len, b_left, b_len)
    return suggest_primers(str(s.seq),
                           primers_pair_ok=primers_pair_ok,
                           opt_t=opt_t, min_t=min_t, max_t=max_t, min_product=100, max_product=500)

def suggest_interval_primers_by_fai(s: pyfaidx.Sequence,
                             a: TranscriptIntersection, b: TranscriptIntersection,
                             opt_t: float = 60.0,
                             min_t: float = 59.9,
                             max_t: float = 65.0) -> PrimerResults:
    return suggest_interval_primers(str(s.seq), s.start, s.end, a, b, opt_t, min_t, max_t)


def extract_one_gene(name: str, annotations: genotations.genomes.Annotations, genome: genomepy.Genome, strand: Strand = Strand.Undefined):
    df: pl.DataFrame = annotations.with_gene_name_in(name).genes() \
        .with_sequences(genome, strand=strand) \
        .annotations_df.select(["start", "end", "strand", "sequence"])
    assert df.shape[0] ==1, f"issue with the gene name as result shape is {df.shape}"
    return df.rows()[0]

def suggest_interval_primers(dna: str, start: int, end: int,
                             a: TranscriptIntersection, b: TranscriptIntersection,
                             opt_t: float = 60.0,
                             min_t: float = 59.9,
                             max_t: float = 65.0, max_product: int = 1000) -> PrimerResults:
    a_left = a.start - start
    a_len = a.end - a.start
    b_left = b.start - end
    b_len = b.end - b.start
    primers_pair_ok = (a_left, a_len, b_left, b_len)
    return suggest_primers(dna, primers_pair_ok=primers_pair_ok, opt_t=opt_t, min_t=min_t, max_t=max_t, min_product=100, max_product=max_product)

def suggest_intervals_intersections_primers(dna: str, start: int, end: int, intervals: list[TranscriptIntersection]):
    results: list[PrimerResults] = []
    for i1 in range(0, len(intervals) -1):
        for i2 in range(i1 + 1, len(intervals)):
            #if abs(intervals[i1].end - intervals[i2].start) <= max_amplified:
            results.append(suggest_interval_primers(dna, start, end, intervals[i1], intervals[i2]))
    return [r for r in results if len(r.get_cleaned_results())>0]




from functools import cached_property

import genomepy

from genotations.genomes import Annotations

ens = genomepy.providers.EnsemblProvider() #instance of ensembl provider to be used for further genome and annotations downloads


def search_assemblies(txt: str):
    """
    just a wrapper to search for existing genome assemblies in Ensembl
    :param txt: search string
    :return: list of found assemblies
    """
    return list(ens.search(txt))


class SpeciesInfo:
    """
    Class to load data from genomepy in an easier way
    """
    assembly: dict
    assembly_name: str
    common_name: str
    species_name: str

    def __init__(self, common_name: str,  assembly_name: str):
        """
        Loads genome and annotations from genomepy in a more organized way
        :param common_name: common name of the species
        :param assembly_name: name of the genome assembly
        """
        assert assembly_name in ens.genomes, "assembly should be in assembly genomes!"
        self.assembly_name = assembly_name
        self.common_name = common_name
        self.assembly = ens.genomes[assembly_name]
        self.species_name = self.assembly["name"]

    @cached_property
    def genome(self):
        """
        Downloads the genome from Ensembl,
        NOTE: this property is cached, can be used in a lazy way!
        :return:
        """
        print("Downloading the genome with annotations from Ensembl, this may take a while. The results are cached")
        genome = genomepy.install_genome(self.assembly_name, "ensembl", annotation=True)
        return genome

    @cached_property
    def annotations(self) -> Annotations:
        """
        Annotation class that is in fact GTF loaded to polars,
        NOTE: if the genome is not downloaded, also starts the download
        :return: Annotation class instance for chained calls
        """
        return Annotations(self.genome.annotation_gtf_file)

mouse = SpeciesInfo("Mouse", "GRCm39") # used for faster access to common mouse genome
human = SpeciesInfo("Human", "GRCh38.p13") # used for faster access to common human genome

species: dict[str, SpeciesInfo] = {
    "mouse": mouse,
    "human": human
}